-- addon constants
local MODNAME   = "BrokerRaidFinder"
local FULLNAME  = "Broker: Raid Finder"
local SHORTNAME = "Raid Finder"
local DIRNAME   = "Broker_RaidFinder"

-- local functions
local strlower          = strlower
local strfind           = strfind
local strupper          = strupper
local string_format     = string.format
local tinsert           = table.insert
local table_sort        = table.sort
local pairs             = pairs
local time              = time
local floor             = floor
local gsub              = gsub

local GetNumRaidMembers  = _G.GetNumRaidMembers
local UnitName           = _G.UnitName
local GetNumPartyMembers = _G.GetNumPartyMembers

-- setup libs
local LibStub   = LibStub
local LDB       = LibStub:GetLibrary("LibDataBroker-1.1")

-- coloring tools
local Crayon	= LibStub:GetLibrary("LibCrayon-3.0")
-- tooltip library
local QT		= LibStub:GetLibrary("LibQTip-1.0")

-- get translations
local L             = LibStub:GetLibrary("AceLocale-3.0"):GetLocale( MODNAME )
local LibBabbleZone = LibStub:GetLibrary("LibBabble-Zone-3.0"):GetLookupTable()

-- config libraries
local AceConfig 		= LibStub:GetLibrary("AceConfig-3.0")
local AceConfigReg 		= LibStub:GetLibrary("AceConfigRegistry-3.0")
local AceConfigDialog	= LibStub:GetLibrary("AceConfigDialog-3.0")

-- utilities
local function Red( text )       return "|cffff0000" .. text .. "|r" end
local function Green( text )     return "|cff00ff00" .. text .. "|r" end
local function Yellow( text )    return "|cffffff00" .. text .. "|r" end
local function Magenta( text )   return "|cffff00ff" .. text .. "|r" end
local function White( text )     return "|cffffffff" .. text .. "|r" end
local function Brownish( text )  return "|cffeda55f" .. text .. "|r" end

local function Orange( text )    return "|cffe54100" .. text .. "|r" end
local function Blueish( text )   return "|cff04adcb" .. text .. "|r" end

local function GrayOut( text )   return "|cff888888" .. text .. "|r" end

local function GetArgs(str, pattern)
   local ret = {}
   local pos=0
   
   while true do
     local word
     _, pos, word=string.find(str, pattern, pos+1)
	 
     if not word then
       break
     end
	 
     word = string.lower(word)
     table.insert(ret, word)
   end
   
   return ret
end

local function format_time(stamp)
	local days    = floor(stamp/86400)
	local hours   = floor((stamp - days * 86400) / 3600)
	local minutes = floor((stamp - days * 86400 - hours * 3600) / 60)
	local seconds = floor((stamp - days * 86400 - hours * 3600 - minutes * 60))

	return string_format("%02d:%02d:%02d", hours, minutes, seconds)
end

-- for sorted keyword list we only care that the shortest keywords (which are more likely to produce a match) are first
local function keyword_lt(a, b) 
	return a:len() < b:len() 
end

-- addon and locals
BrokerRaidFinder = LibStub:GetLibrary("AceAddon-3.0"):NewAddon(MODNAME, "AceEvent-3.0", "AceConsole-3.0", "AceTimer-3.0")
local BrokerRaidFinder = _G.BrokerRaidFinder

LibStub:GetLibrary('LibWho-2.0'):Embed(BrokerRaidFinder)

local tooltip

local function CreateTooltip( self )
	tooltip = QT:Acquire( MODNAME.."TT", 3 )
	tooltip:Hide()
	tooltip:Clear()
	tooltip:SetScale(1)
		
	BrokerRaidFinder:DrawTooltip()

	tooltip:SetAutoHideDelay( .1, self )
	tooltip:EnableMouse()
	tooltip:SmartAnchorTo( self )
	tooltip:Show()
end

local function RemoveTooltip()
	if tooltip then
		tooltip:Hide()
		QT:Release(tooltip)
		tooltip = nil
	end
end

local dataobj  = LDB:NewDataObject(MODNAME, {
	type	= "data source",
	icon	= "Interface\\Addons\\"..DIRNAME.."\\icon",
	label	= MODNAME,
	text	= SHORTNAME,
	OnClick = function(clickedframe, button) 
		BrokerRaidFinder:OnClick(button) 
	end,
	OnEnter = function ( self )
		CreateTooltip(self)
	end,
	OnLeave = function()
		-- RemoveTooltip()
		return
	end,
})

local function BRF_ToggleInstance(cell, instance, button)
	BrokerRaidFinder:ToggleInstance(instance)
	BrokerRaidFinder:DrawTooltip()
	tooltip:Show()
end

--  minimap button
do
	local dragMode = nil
	
	local function moveButton(self)
		if dragMode == "free" then
			local centerX, centerY = Minimap:GetCenter()
			local x, y = GetCursorPosition()
			x, y = x / self:GetEffectiveScale() - centerX, y / self:GetEffectiveScale() - centerY
			self:ClearAllPoints()
			self:SetPoint("CENTER", x, y)
		else
			local centerX, centerY = Minimap:GetCenter()
			local x, y = GetCursorPosition()
			x, y = x / self:GetEffectiveScale() - centerX, y / self:GetEffectiveScale() - centerY
			centerX, centerY = abs(x), abs(y)
			centerX, centerY = (centerX / sqrt(centerX^2 + centerY^2)) * 80, (centerY / sqrt(centerX^2 + centerY^2)) * 80
			centerX = x < 0 and -centerX or centerX
			centerY = y < 0 and -centerY or centerY
			self:ClearAllPoints()
			self:SetPoint("CENTER", centerX, centerY)
		end
	end

	local mm_button = CreateFrame("Button", MODNAME .. "_MinimapButton", Minimap)
	mm_button:SetHeight(32)
	mm_button:SetWidth(32)
	mm_button:SetFrameStrata("HIGH")
	mm_button:SetPoint("CENTER", -77.44, -20.06)
	mm_button:SetMovable(true)
	mm_button:SetUserPlaced(true)
	mm_button:EnableMouse(true)
	mm_button:RegisterForClicks("LeftButtonUp", "RightButtonUp")
	mm_button:SetHighlightTexture("Interface\\Minimap\\UI-Minimap-ZoomButton-Highlight")
	
	local icon = mm_button:CreateTexture(mm_button:GetName() .. "Icon", "HIGH")
	icon:SetTexture("Interface\\Addons\\"..DIRNAME.."\\icon.tga")
	icon:SetTexCoord(0, 1, 0, 1)
	icon:SetWidth(16)
	icon:SetHeight(16)
	icon:SetPoint("TOPLEFT", mm_button, "TOPLEFT", 8, -7)
	
	mm_button.icon = icon
	
	local overlay = mm_button:CreateTexture(mm_button:GetName() .. "Overlay", "OVERLAY")
	overlay:SetTexture("Interface\\Minimap\\MiniMap-TrackingBorder")
	overlay:SetWidth(53)
	overlay:SetHeight(53)
	overlay:SetPoint("TOPLEFT", mm_button, "TOPLEFT")
	
	mm_button:SetScript("OnClick", function(self, button)
		BrokerRaidFinder:OnClick(button)
	end)
	
	mm_button:SetScript("OnMouseDown", function(self, button)
		if IsShiftKeyDown() and IsControlKeyDown() then
			dragMode = "free"
			self:SetScript("OnUpdate", moveButton)
		elseif IsShiftKeyDown() then
			dragMode = nil
			self:SetScript("OnUpdate", moveButton)
		end
	end)
	
	mm_button:SetScript("OnMouseUp", function(self)
		self:SetScript("OnUpdate", nil)
	end)

	mm_button:SetScript("OnEnter", function(self)
		CreateTooltip(self)
	end)
	
	mm_button:SetScript("OnLeave", function(self)
		RemoveTooltip()
	end)
	
	mm_button:Hide()

	function BrokerRaidFinder:ShowMinimapButton(show)
		if show then
			mm_button:Show()
		else
			mm_button:Hide()
		end
	end

	function BrokerRaidFinder:SetMinimapIcon(active)
		if active then
			mm_button.icon:SetTexture("Interface\\Addons\\"..DIRNAME.."\\icon.tga")
		else
			mm_button.icon:SetTexture("Interface\\Addons\\"..DIRNAME.."\\icondisabled.tga")
		end
	end
end

-- LOCAL CONSTANTS
local GTIMERANGE = 120	
	
-- LOCAL VARIABLES
local options = {}
local defaults = {
	profile = {
		monitoringActive    = true,
		monitorGuild        = true,
		excludeSavedRaids   = false,
		notifyText          = true,
		notifySound         = false,
		notificationSound   = 0,
		notificationTimeout = 5,
		timeFrame           = 60,
		addonCommunication  = false,
		hideHint            = false,
        minimap	            = false,
		monitored           = {},
	},
	faction = {
		matches = {},
	},
	global = {
		lfgKeywords      = brf_DefaultLFGKeywords,
		instanceKeywords = {},
	}
}

local soundFiles = {
	[0] = "Sound\\interface\\iTellMessage.wav",
	[1] = "Sound\\Spells\\BeastCall.wav",
	[2] = "Sound\\Doodad\\BellTollAlliance.wav",
	[3] = "Sound\\Doodad\\BellTollHorde.wav",
	[4] = "Sound\\Doodad\\BellTollNightElf.wav",
	[5] = "Sound\\Doodad\\BellTollTribal.wav",
	[6] = "Sound\\Doodad\\BlastedLandsLightningbolt01Stand-Bolt.wav",
	[7] = "Sound\\Spells\\Bonk1.wav",
	[8] = "Sound\\Spells\\Bonk2.wav",
	[9] = "Sound\\Spells\\Bonk3.wav",
	[10] = "Sound\\Doodad\\TK_Control_Consoles.wav",
	[11] = "Sound\\interface\\DropOnGround.wav",
	[12] = "Sound\\interface\\DwarfExploration.wav",
	[13] = "Sound\\Doodad\\DwarfHorn.wav",
	[14] = "Sound\\interface\\Error.wav",
	[15] = "Sound\\Doodad\\Firecrackers_ThrownExplode.wav",
	[16] = "Sound\\Doodad\\G_GongTroll01.wav",
	[17] = "Sound\\Events\\GuldanCheers.wav",
	[18] = "Sound\\Doodad\\KharazahnBellToll.wav",
	[19] = "Sound\\Spells\\KillCommand.wav",
	[20] = "Sound\\interface\\MapPing.wav",
	[21] = "Sound\\Doodad\\G_Mortar.wav",
	[22] = "Sound\\interface\\RaidWarning.wav",
	[23] = "Sound\\interface\\ReadyCheck.wav",
	[24] = "Sound\\Doodad\\BE_ScryingOrb_Explode.wav",
	[25] = "Sound\\Doodad\\Sizzle.wav",
	[26] = "Sound\\Spells\\ThrowWater.wav",
	[27] = "Sound\\interface\\UnsheathMetal.wav",
	[28] = "Sound\\interface\\UnsheathShield.wav",
	[29] = "Sound\\Event Sounds\\Wisp\\WispPissed1.wav",
	[30] = "Sound\\Event Sounds\\Wisp\\WispReady1.wav",
	[31] = "Sound\\Event Sounds\\Wisp\\WispWhat1.wav",
	[32] = "Sound\\Event Sounds\\Wisp\\WispYes1.wav",
}

local soundNames = {
	[0] = L["Default"].." (Tell Message)",
	[1] = "Beast Call",
	[2] = "Bell Toll Alliance",
	[3] = "Bell Toll Horde",
	[4] = "Bell Toll Night Elf",
	[5] = "Bell Toll Tribal",
	[6] = "Bolt",
	[7] = "Bonk 1",
	[8] = "Bonk 2",
	[9] = "Bonk 3",
	[10] = "Control Consoles",
	[11] = "Drop On Ground",
	[12] = "Dwarf Exploration",
	[13] = "Dwarf Horn",
	[14] = "Error",
	[15] = "Firecrackers",
	[16] = "Gong Troll",
	[17] = "Guldan Cheers",
	[18] = "Kharazahn Bell Toll",
	[19] = "Kill Command",
	[20] = "Map Ping",
	[21] = "Mortar",
	[22] = "Raid Warning",
	[23] = "Ready Check",
	[24] = "ScryingOrb Explode",
	[25] = "Sizzle",
	[26] = "ThrowWater",
	[27] = "Unsheath Metal",
	[28] = "Unsheath Shield",
	[29] = "Wisp Pissed",
	[30] = "Wisp Ready",
	[31] = "Wisp What",
	[32] = "Wisp Yes",
}

-- log window stuff
local MAX_LIST_SIZE       = 10
local LISTENTRYHEIGHT     = 20

local FILTERINSTOPT_ALL   = "Show All"

local FILTERSRCOPT_ALL    = 0
local FILTERSRCOPT_SELF   = 1
local FILTERSRCOPT_ALTS   = 2
local FILTERSRCOPT_REMOTE = 3

local filterInstanceBtnInfo = { func = function(self) BrokerRaidFinder:HandleInstanceFilter(self.value); UIDropDownMenu_SetSelectedValue(BrokerRaidFinder.logfilterinstance, self.value) end } 
local filterSourceBtnInfo   = { func = function(self) BrokerRaidFinder:HandleSourceFilter(self.value); UIDropDownMenu_SetSelectedValue(BrokerRaidFinder.logfiltersource, self.value) end }

local function InitInstanceFilter()
	filterInstanceBtnInfo.text = L["Show All"]
	filterInstanceBtnInfo.value = FILTERINSTOPT_ALL
	filterInstanceBtnInfo.checked = BrokerRaidFinder.instancefilter == FILTERINSTOPT_ALL
	UIDropDownMenu_AddButton(filterInstanceBtnInfo)
	
	-- populate with currently monitored instances
	local monitored = {}
	
	for instance, _ in pairs(BrokerRaidFinder.db.profile.monitored) do
		tinsert(monitored, instance)
	end
	
	table_sort(monitored, function(a, b) return LibBabbleZone[a] < LibBabbleZone[b]; end)
	
	for i = 1, #monitored do
		filterInstanceBtnInfo.text =  LibBabbleZone[monitored[i]]
		filterInstanceBtnInfo.value = monitored[i]
		filterInstanceBtnInfo.checked = BrokerRaidFinder.instancefilter == monitored[i]
		UIDropDownMenu_AddButton(filterInstanceBtnInfo)
	end

	UIDropDownMenu_SetSelectedValue(BrokerRaidFinder.logfilterinstance, BrokerRaidFinder.instancefilter)
end

local function InitSourceFilter()
	filterSourceBtnInfo.text = L["Show All"]
	filterSourceBtnInfo.value = FILTERSRCOPT_ALL
	filterSourceBtnInfo.checked = BrokerRaidFinder.sourcefilter == FILTERSRCOPT_ALL
	UIDropDownMenu_AddButton(filterSourceBtnInfo)
	
	filterSourceBtnInfo.text = L["Self"]
	filterSourceBtnInfo.value = FILTERSRCOPT_SELF
	filterSourceBtnInfo.checked = BrokerRaidFinder.sourcefilter == FILTERSRCOPT_SELF
	UIDropDownMenu_AddButton(filterSourceBtnInfo)
	
	filterSourceBtnInfo.text = L["Alts"]
	filterSourceBtnInfo.value = FILTERSRCOPT_ALTS
	filterSourceBtnInfo.checked = BrokerRaidFinder.sourcefilter == FILTERSRCOPT_ALTS
	UIDropDownMenu_AddButton(filterSourceBtnInfo)
	
	filterSourceBtnInfo.text = L["Remote"]
	filterSourceBtnInfo.value = FILTERSRCOPT_REMOTE
	filterSourceBtnInfo.checked = BrokerRaidFinder.sourcefilter == FILTERSRCOPT_REMOTE
	UIDropDownMenu_AddButton(filterSourceBtnInfo)
	
	UIDropDownMenu_SetSelectedValue(BrokerRaidFinder.logfiltersource, BrokerRaidFinder.sourcefilter)
end

local function MatchListUpdate()
	local frame = BrokerRaidFinder.matchlistframe
	local entries = frame.entries
	local numEntries = #BrokerRaidFinder.filtered_matches
	local listSize = min(numEntries, MAX_LIST_SIZE)

	local scrollFrame = frame.scrollFrame
	
	local offset = FauxScrollFrame_GetOffset(scrollFrame)
	FauxScrollFrame_Update(scrollFrame, numEntries, listSize, LISTENTRYHEIGHT)

	-- if the list gets shortened so much that all entries are below the current offset then adjust the offset
	if offset + MAX_LIST_SIZE > numEntries then
		offset = numEntries - MAX_LIST_SIZE
		if offset < 0 then
			offset = 0
		end
		FauxScrollFrame_SetOffset(scrollFrame, offset)
	end	
	
	for i = 1, MAX_LIST_SIZE do
		local index = i + offset
		local entry = entries[i]

		if index <= numEntries then
			local data = BrokerRaidFinder.filtered_matches[index]
			entry.timeframe.text:SetText(format_time(data.timestamp+BrokerRaidFinder.time_offset))
			entry.authorframe.text:SetText(BrokerRaidFinder:ColorizeChar(tostring(data.player)))
			entry.msgframe.text:SetText(LibBabbleZone[data.instance])
			
			entry.data = data
			
			if BrokerRaidFinder.selected_match == data then
				entry:SetBackdropColor(1, 1, 1, 1)
			else
				entry:SetBackdropColor(0, 0, 0, .33)
			end
			
			entry:Show()
		else
			entry:Hide()
		end
		
		if scrollFrame:IsShown() then
			entry:SetPoint("TOPRIGHT", frame, "TOPRIGHT", -20, -1*(i-1)*LISTENTRYHEIGHT - 3)
		else
			entry:SetPoint("TOPRIGHT", frame, "TOPRIGHT", -3, -1*(i-1)*LISTENTRYHEIGHT - 3)
		end		
	end
end

-- blizzard channels to monitor
local monitored_channels = {
	[1] =  true, -- General
	[2] =  true, -- Trade
	[4] =  true, -- LFG
}

-- infrastcructure
function BrokerRaidFinder:SetupOptions()	
    options = {
		handler = BrokerRaidFinder,
		type = 'group',
		args = {
			raids = {
				type = 'group',
				name = L["Raids"],
				desc = L["Set up which raids you will monitor."],
				order = 10,
				args = {
					-- filled dynamically
				},
			},			
			keywords = {
				type = 'group',
				name = L["Keywords"],
				desc = L["Set up keywords for each instance."],
				order = 20,
				args = {
					lfgkeywords = {
						type = "input",
						name = L["LFG Keywords"],
						desc = L["Comma separated list of keywords indicating someone is looking for players for a raid."],
						order = 1,
						multiline = 4,
						get = function()
								return self.db.global.lfgKeywords
							end,
						set = function(info, key)
							self.db.global.lfgKeywords = key
							self:UpdateLFGKeywords()
						end,
					},
					-- restore default keywords button
					defaultlfgkeywords = {
						type = "execute",
						name = L["Default"],
						desc = L["Revert to default LFG keywords"],
						order = 2,
						func = function()
							self.db.global.lfgKeywords = L["LFGDefaultKeywords"]
							self:UpdateLFGKeywords()
						end,	
					}
					-- raid keywords filled dynamically
				},
			},			
			monitoring = {
				type = 'group',
				name = L["Monitoring"],
				desc = L["Set up monitoring parmaters for addon."],
				order = 30,
				args = {
					guild = {
						type = 'toggle',
						name = L["Guild Chat"],
						desc = L["Monitor guild chat."],
						order = 1,
						get = function() return self.db.profile.monitorGuild end,
						set = function() 
							self:ToggleMonitorGuild()
						end,
					},
					saved = {
						type = 'toggle',
						name = L["Exclude Saved Raids"],
						desc = L["Exclude raids you are currently saved to."],
						order = 2,
						get = function() return self.db.profile.excludeSavedRaids end,
						set = function() 
							self:ToggleSavedRaids()
						end,
					},
					timeframe = {
						type = 'range',
						name = L["Time Frame"],
						desc = L["Set up how many minutes the log will reach back."],
						order = 3,
						get = function() return self.db.profile.timeFrame end,
						set = function(info, v) 
							self.db.profile.timeFrame = v
							self.tainted_data = true
							self:Update()
						end,
						min = 0,
						max = GTIMERANGE,
						step = 1,
					},
				},
			},			
			notifications = {
				type = 'group',
				name = L["Notifications"],
				order = 30,
				desc = L["Set up notifications when addon finds a match."],
				args = {
					text = {
						type = 'toggle',
						name = L["Text Alert"],
						desc = L["Show text message when addon finds a match."],
						order = 1,
						get = function() return self.db.profile.notifyText end,
						set = function() 
							self.db.profile.notifyText = not self.db.profile.notifyText
						end,
					},
					sound = {
						type = 'toggle',
						name = L["Sound Alert"],
						desc = L["Play sound when addon finds a match."],
						order = 2,
						get = function() return self.db.profile.notifySound end,
						set = function() 
							self.db.profile.notifySound = not self.db.profile.notifySound
						end,
					},
					notificationSound = { 
						type = 'select',
						name = L["Notification Sound"],
						desc = L["Choose sound to be played on notifications."],
						order = 3,
						get = function() return self.db.profile.notificationSound end,	
						set = function(info, key) 
							self.db.profile.notificationSound = key
						end,
						values = soundNames,
					},
					playSound = { 
						type = 'execute',
						name = L["Play Sound"],
						desc = L["Play selected notification sound."],
						order = 4,
						func = function() self:PlaySoundFile() end,	
					},
					timeout = {
						type = 'range',
						name = L["Timeout"],
						desc = L["Set notification timeout. You will not be notified about matches of a single player for the same instance more than once during the timeout duration."],
						order = 5,
						get = function() return self.db.profile.notificationTimeout end,
						set = function(info, v) 
							self.db.profile.notificationTimeout = v
						end,
						min = 0,
						max = GTIMERANGE,
						step = 1,
						order = 1
					},
				},
			},
			active = { 
				type = "toggle",
				name = L["Monitoring Active"],
				desc = L["Activate/Deactivate the monitoring."],
	            order = 1,
				get = function() return self.db.profile.monitoringActive end,	
				set = function(info, key)
					self:ToggleMonitoring()
				end,
			},
			addonComm  = {
				type = "toggle",
				name = L["Addon Communication"],
				desc = L["Toggle whether or not the addon shall sync with addons of other players. This is restricted to (mutual) friends and guild members."],
				order = 2,
				get = function() return self.db.profile.addonCommunication end,
				set = function() 
					self:ToggleAddonCommunication()
				end,
			},
			minimap = {
				type = 'toggle',
				name = L["Minimap Button"],
				desc = L["Show Minimap Button"],
				order = 3,
				get  = function() return self.db.profile.minimap end,
				set  = function()
					self.db.profile.minimap = not self.db.profile.minimap
					
					BrokerRaidFinder:ShowMinimapButton(self.db.profile.minimap)
				end,
			},
			hint = {
				type = 'toggle',
				name = L["Hide Hint"],
				desc = L["Hide usage hint in tooltip"],
				order = 4,
				get  = function() return self.db.profile.hideHint end,
				set  = function()
					self.db.profile.hideHint = not self.db.profile.hideHint; 
				end,
			},
		},
	}
	
	self:InsertRaidOptions()
end

function BrokerRaidFinder:InsertRaidOptions()
	for extension, instances in pairs(brf_Instances) do
		
		-- preface raids by extension name
		options.args.raids.args[brf_Extensions[extension]] = {
			type = "description",
			name = L[brf_Extensions[extension]],
			order = extension*100,
		}
		
		options.args.keywords.args[brf_Extensions[extension]] = {
			type = "description",
			name = L[brf_Extensions[extension]],
			order = extension*100,
		}
		
		for index, instance in pairs(instances) do
			-- raid monitoring
			options.args.raids.args[instance] = {
				type = 'toggle',
				name = LibBabbleZone[instance],
				desc = L["Monitor chat for "] .. LibBabbleZone[instance] .. ".",
				order = extension*100+10*index,
				get = function() return self.db.profile.monitored[instance] == true end,
				set = function() 
					self:ToggleInstance(instance)
				end,
			}
			-- raid keywords
			options.args.keywords.args[instance] = {
				type = "input",
				name = LibBabbleZone[instance],
				desc = L["Keywords for "] .. LibBabbleZone[instance] .. ".",
				order = extension*100+10*index,
				multiline = 4,
				get = function() return self.db.global.instanceKeywords[instance] end,
				set = function(info, key)
					self.db.global.instanceKeywords[instance] = key
					self:UpdateInstanceKeywords(instance)
				end,
			}
			
			-- restore default keywords button
			options.args.keywords.args[instance.."-default"] = {
				type = "execute",
				name = L["Default"],
				desc = L["Revert to default keywords for "] .. LibBabbleZone[instance] .. ".",
				order = extension*100+10*index+1,
				func = function()
					self:SetDefaultKeywords(instance)
					self:UpdateInstanceKeywords(instance)
				end,	
			}
		end
	end	
end

function BrokerRaidFinder:OnInitialize()
	-- addon variables
	self.player_saved_raids = {}
	self.notification_timeouts = {}
	
	-- debugging
	self.debug = false
	
	-- unit information
	self.raidNames   = {}
	self.PlayerName  = UnitName("player")
	self.playerName  = strlower(self.PlayerName)

	self.classCache  = {}
	
	-- offset to local time
	local gHour = GetGameTime()
	self.time_offset = (tonumber(gHour) - date("!%H")) * 3600
	
	-- processed data
	self.tainted_data      = true
	self.matched_players   = 0
	self.matched_instances = 0
	self.matches_latest    = {}
	self.filtered_matches  = {}
	
	-- keyword data
	self.lfgKeywordList       = {}
	self.instanceKeywordLists = {}
	
	-- insert default keywords
	self:SetDefaultInstanceKeywords(true)
		
	-- options
	self.db = LibStub:GetLibrary("AceDB-3.0"):New(MODNAME.."_DB", defaults, "Default")
	
	self:SetupOptions()

	-- profile support
	options.args.profile = LibStub:GetLibrary("AceDBOptions-3.0"):GetOptionsTable(self.db)
	self.db.RegisterCallback(self, "OnProfileChanged", "OnProfileChanged")
	self.db.RegisterCallback(self, "OnProfileCopied",  "OnProfileChanged")
	self.db.RegisterCallback(self, "OnProfileReset",   "OnProfileChanged")

	AceConfigReg:RegisterOptionsTable(FULLNAME, options)
	AceConfigDialog:AddToBlizOptions(FULLNAME)

	self:RegisterChatCommand("braidfinder", "ChatCommand")
    self:RegisterChatCommand("brfind",      "ChatCommand")
	
	-- log window setup
	local backdrop = {
		-- path to the background texture
		bgFile = "Interface\\DialogFrame\\UI-DialogBox-Background", 
		-- path to the border texture
		edgeFile = "Interface\\TutorialFrame\\TutorialFrameBorder",
		-- true to repeat the background texture to fill the frame, false to scale it
		tile = true,
		-- size (width or height) of the square repeating background tiles (in pixels)
		tileSize = 32,
		-- thickness of edge segments and square size of edge corners (in pixels)
		edgeSize = 28,
		-- distance from the edges of the frame to those of the background texture (in pixels)
		insets = {
			left   = 5,
			right  = 2,
			top    = 20,
			bottom = 5
		}
	}
	
	self.logwindow = CreateFrame("Frame", MODNAME.."LogWindow", UIParent)
	self.logwindow:SetFrameStrata("DIALOG")
	self.logwindow:CreateTitleRegion():SetAllPoints()
	self.logwindow:SetToplevel(true)
	self.logwindow:EnableMouse(true)
	self.logwindow:SetMovable(true)
	self.logwindow:SetUserPlaced(true)
	self.logwindow:SetHeight(370)
	self.logwindow:SetWidth(500)
	self.logwindow:SetBackdrop(backdrop)
	self.logwindow:SetPoint("CENTER", UIParent, "CENTER", 0, 0)
	self.logwindow:RegisterForDrag("LeftButton")
	
	tinsert(UISpecialFrames, self.logwindow:GetName())
	
	local titlebar = self.logwindow:CreateTitleRegion()
	titlebar:SetPoint("TOPRIGHT", -5, 0)
	titlebar:SetPoint("TOPLEFT", 5, 0)
	titlebar:SetHeight(28)

	self.logwindow.title = self.logwindow:CreateFontString(nil, "OVERLAY", "GameFontHighlight")
	self.logwindow.title:SetPoint("CENTER", titlebar, "CENTER", 0, 8)
	self.logwindow.title:SetText(FULLNAME .. " - " .. L["Log Window"])
	
	local button = CreateFrame("Button", nil, self.logwindow, "UIPanelCloseButton")
	button:SetPoint("TOPRIGHT", 5, 6)
	
	-- label for instance filter
	local frame = CreateFrame("Frame", nil, self.logwindow)
	frame:SetHeight(20)
	frame:SetWidth(85)
	frame:SetPoint("TOPLEFT", self.logwindow, "TOPLEFT", 15, -35)
	
	local text = frame:CreateFontString(nil, "OVERLAY", "GameFontNormal")
	text:SetPoint("LEFT", frame, "LEFT")
	text:SetText(L["Show Instance:"])
	
	-- combobox for instance filter
	self.instancefilter = FILTERINSTOPT_ALL
	
	self.logfilterinstance = CreateFrame("Frame", MODNAME.."InstanceFilter", self.logwindow, "UIDropDownMenuTemplate")
	
	self.logfilterinstance:ClearAllPoints()
	self.logfilterinstance:SetPoint("TOPLEFT", self.logwindow, "TOPLEFT", 100, -30)
	
	UIDropDownMenu_Initialize(self.logfilterinstance, InitInstanceFilter)
	UIDropDownMenu_SetWidth(self.logfilterinstance, 100);
	UIDropDownMenu_SetButtonWidth(self.logfilterinstance, 124)
	UIDropDownMenu_JustifyText(self.logfilterinstance, "LEFT")
 
	-- label for source filter
	local frame = CreateFrame("Frame", nil, self.logwindow)
	frame:SetHeight(20)
	frame:SetWidth(85)
	frame:SetPoint("TOPLEFT", self.logwindow, "TOPLEFT", 260, -35)
	
	local text = frame:CreateFontString(nil, "OVERLAY", "GameFontNormal")
	text:SetPoint("LEFT", frame, "LEFT")
	text:SetText(L["Show Source:"])

	-- combobox for source filter
	self.sourcefilter = FILTERSRCOPT_ALL
	
	self.logfiltersource = CreateFrame("Frame", MODNAME.."SourceFilter", self.logwindow, "UIDropDownMenuTemplate")
	
	self.logfiltersource:ClearAllPoints()
	self.logfiltersource:SetPoint("TOPLEFT", self.logwindow, "TOPLEFT", 355, -30)
	
	UIDropDownMenu_Initialize(self.logfiltersource, InitSourceFilter)
	UIDropDownMenu_SetWidth(self.logfiltersource, 100);
	UIDropDownMenu_SetButtonWidth(self.logfiltersource, 124)
	UIDropDownMenu_JustifyText(self.logfiltersource, "LEFT")
	
	-- THE list
	backdrop = {
		-- path to the background texture
		bgFile = "Interface\\DialogFrame\\UI-DialogBox-Background-Dark", 
		-- path to the border texture
		edgeFile = "Interface\\Tooltips\\UI-Tooltip-Border",
		-- true to repeat the background texture to fill the frame, false to scale it
		tile = true,
		-- size (width or height) of the square repeating background tiles (in pixels)
		tileSize = 16,
		-- thickness of edge segments and square size of edge corners (in pixels)
		edgeSize = 16,
		-- distance from the edges of the frame to those of the background texture (in pixels)
		insets = {
			left   = 5,
			right  = 5,
			top    = 5,
			bottom = 5
		}
	}
	frame = CreateFrame('Frame', self.logwindow:GetName() .. 'FSList', self.logwindow)
	frame:SetHeight(MAX_LIST_SIZE*LISTENTRYHEIGHT+6)
	frame:SetWidth(480)
	frame:SetBackdrop(backdrop)
	frame:SetPoint("TOPLEFT", self.logwindow, "TOPLEFT", 10, -70)

	-- list entries
	frame.entries = {}

	backdrop = {
		-- path to the background texture
		bgFile = "Interface\\QuestFrame\\UI-QuestTitleHighlight", 
		-- true to repeat the background texture to fill the frame, false to scale it
		tile = false,
		-- distance from the edges of the frame to those of the background texture (in pixels)
		insets = {
			left   = 1,
			right  = 1,
			top    = 1,
			bottom = 1
		}
	}
	
	for i=1, MAX_LIST_SIZE do
		local entry = CreateFrame('Button', frame:GetName() .. 'Entry' .. tostring(i), frame)
		entry:SetHeight(LISTENTRYHEIGHT)
		entry:SetPoint("TOPLEFT",  frame, "TOPLEFT",   3, -1*(i-1)*LISTENTRYHEIGHT - 3)
		entry:SetPoint("TOPRIGHT", frame, "TOPRIGHT", -3, -1*(i-1)*LISTENTRYHEIGHT - 3)
		entry:SetBackdrop(backdrop)
		entry:SetBackdropColor(0, 0, 0, .33)
		-- entry:EnableMouse(true)
		entry:SetScript('OnEnter', function(self) self:SetBackdropColor(1, 1, 1, 1); if self.data then GameTooltip:SetOwner(self, 'ANCHOR_CURSOR'); GameTooltip:SetText(self.data.message); end; end)
		entry:SetScript('OnLeave', function(self) if self.data ~= BrokerRaidFinder.selected_match then self:SetBackdropColor(0, 0, 0, .33); end; GameTooltip:Hide(); end)
		entry:SetScript('OnClick', function(self, button, down) BrokerRaidFinder:SetSelectedEntry(self) end)
		
		frame.entries[i] = entry
		
		-- subframes holding text for time, author and message
		entry.timeframe = CreateFrame('Frame', entry:GetName() .. '1', entry)
		entry.timeframe:SetHeight(LISTENTRYHEIGHT)
		entry.timeframe:SetWidth(60)
		entry.timeframe:SetPoint("TOPLEFT", entry, "TOPLEFT", 5, 0)
		entry.timeframe.text = entry.timeframe:CreateFontString(nil, "OVERLAY", "GameFontNormalSmall")
		entry.timeframe.text:SetAllPoints()
		entry.timeframe.text:SetJustifyH("LEFT")
		
		entry.authorframe = CreateFrame('Frame', entry:GetName() .. '1', entry)
		entry.authorframe:SetHeight(LISTENTRYHEIGHT)
		entry.authorframe:SetWidth(100)
		entry.authorframe:SetPoint("TOPLEFT", entry.timeframe, "TOPRIGHT", 5, 0)
		entry.authorframe.text = entry.authorframe:CreateFontString(nil, "OVERLAY", "GameFontNormalSmall")
		entry.authorframe.text:SetAllPoints()
		entry.authorframe.text:SetJustifyH("LEFT")
		
		entry.msgframe = CreateFrame('Frame', entry:GetName() .. '1', entry)
		entry.msgframe:SetHeight(LISTENTRYHEIGHT)
		entry.msgframe:SetPoint("TOPLEFT", entry.authorframe, "TOPRIGHT", 5, 0)
		entry.msgframe:SetPoint("TOPRIGHT", entry, "TOPRIGHT")
		entry.msgframe.text = entry.msgframe:CreateFontString(nil, "OVERLAY", "GameFontNormalSmall")
		entry.msgframe.text:SetAllPoints()
		entry.msgframe.text:SetJustifyH("LEFT")
	end
	
	-- the scrollframe
	local scroll = CreateFrame('ScrollFrame', frame:GetName() .. 'ScrollFrame', frame, 'FauxScrollFrameTemplate')
	scroll:SetPoint('TOPLEFT', 5, -4)
	scroll:SetPoint('BOTTOMRIGHT', -26, 3)
	scroll:SetScript('OnVerticalScroll', function(self, arg1)
		FauxScrollFrame_OnVerticalScroll(self, arg1, LISTENTRYHEIGHT, function() MatchListUpdate() end)
	end)
	frame.scrollFrame = scroll

	frame:SetScript('OnShow', function(self) MatchListUpdate() end)
	
	self.matchlistframe = frame
	
	-- frame for the whole entry including message
	backdrop = {
		-- path to the background texture
		bgFile = "Interface\\DialogFrame\\UI-DialogBox-Background-Dark", 
		-- path to the border texture
		edgeFile = "Interface\\Tooltips\\UI-Tooltip-Border",
		-- true to repeat the background texture to fill the frame, false to scale it
		tile = true,
		-- size (width or height) of the square repeating background tiles (in pixels)
		tileSize = 16,
		-- thickness of edge segments and square size of edge corners (in pixels)
		edgeSize = 16,
		-- distance from the edges of the frame to those of the background texture (in pixels)
		insets = {
			left   = 5,
			right  = 5,
			top    = 5,
			bottom = 5
		}
	}
	local frame2 = CreateFrame('Frame', self.logwindow:GetName() .. 'Message', self.logwindow)
	frame2:SetHeight(40)
	frame2:SetWidth(480)
	frame2:SetBackdrop(backdrop)
	frame2:SetPoint("TOPLEFT", self.logwindow, "TOPLEFT", 10, -285)
	
	frame2.text = frame2:CreateFontString(nil, "OVERLAY", "GameFontNormalSmall")
	frame2.text:SetAllPoints()
	frame2.text:SetPoint("TOPLEFT",     frame2, "TOPLEFT",      5, -5)
	frame2.text:SetPoint("BOTTOMRIGHT", frame2, "BOTTOMRIGHT", -5,  5)
	frame2.text:SetJustifyH("LEFT")
	frame2.text:SetJustifyV("TOP")
	frame2.text:SetWordWrap(true)
	frame2.text:SetNonSpaceWrap(false) 
	
	self.messageframe = frame2
	
	-- whisper button
	local button = CreateFrame("Button", nil, self.logwindow, "UIPanelButtonTemplate")
	button:SetHeight(20)
	button:SetWidth(80)
	button:SetText("Whisper")
	button:SetScript("OnClick", function(self, button, down) BrokerRaidFinder:WhisperAuthor() end)
	button:SetPoint("TOPLEFT", self.logwindow, "TOPLEFT", 210, -335)
	
	self.whisperbutton = button
end

function BrokerRaidFinder:OnEnable()
	-- redo: offset to local time
	-- seems on initialize the game time is not always set properly
	local gHour = GetGameTime()
	self.time_offset = (tonumber(gHour) - date("!%H")) * 3600
	
	self:SetupEventHandlers()
	
	-- register raid info events
	self:RegisterEvent("UPDATE_INSTANCE_INFO")
	self:RegisterEvent("RAID_INSTANCE_WELCOME")
	self:RegisterEvent("CHAT_MSG_SYSTEM")
	
	-- register raid unit events
	self:RegisterEvent("PARTY_CONVERTED_TO_RAID", "UpdateRaidUnits")
	self:RegisterEvent("PARTY_MEMBERS_CHANGED", "UpdateRaidUnits")
	
	-- wholib query callback
	-- self:RegisterCallback('WHOLIB_QUERY_RESULT', 'UnitInfoCallback')
	
	-- timer for cyclic cleaning up of history
	self.timer = self:ScheduleRepeatingTimer("ManageHistory", 60)
	
	-- update keyword data
	self:UpdateAllKeywords()
	
	-- update cached unit data
	self:UpdateRaidUnits()
	
	self:ShowMinimapButton(self.db.profile.minimap)
	self:UpdateIcon()
	self:Update()
	
	-- request info about saved instances
	RequestRaidInfo()
	
	-- setup comm
	self:SetupCommunication()
end

function BrokerRaidFinder:OnDisable()
	-- shut down comm
	self:ShutdownCommunication()

	-- unregister chat events
	self:UnregisterEvent("CHAT_MSG_CHANNEL")
	self:UnregisterEvent("CHAT_MSG_GUILD")
	self:UnregisterEvent("CHAT_MSG_OFFICER")	
	
	-- unregister raid info events
	self:UnregisterEvent("UPDATE_INSTANCE_INFO")
	self:UnregisterEvent("RAID_INSTANCE_WELCOME")
	self:UnregisterEvent("CHAT_MSG_SYSTEM")

	-- unregister raid unit events
	self:UnregisterEvent("PARTY_CONVERTED_TO_RAID")
	self:UnregisterEvent("PARTY_MEMBERS_CHANGED")
	
	-- cancel timer for cyclic cleaning up of history
	self:CancelTimer(self.timer)
end

function BrokerRaidFinder:OnProfileChanged(event, database, newProfileKey)
	self.db.profile = database.profile
	
	self:SetupEventHandlers()
	
	self.tainted_data = true

	self:UpdateIcon()
	self:Update()
end

function BrokerRaidFinder:ChatCommand(input)
    if input then  
		args = GetArgs(input, "^ *([^%s]+) *")
		
		BrokerRaidFinder:TriggerAction(args[1], args)
	else
		BrokerRaidFinder:TriggerAction("help")
	end
end

function BrokerRaidFinder:OnClick(button)
	if ( button == "RightButton" ) then 
		if IsShiftKeyDown() then
			-- unused
		elseif IsControlKeyDown() then
			-- unused
		elseif IsAltKeyDown() then
			-- unused
		else
			-- open options menu
			BrokerRaidFinder:TriggerAction("menu")
		end
	elseif ( button == "LeftButton" ) then 
		if IsShiftKeyDown() then
			-- unused
		elseif IsControlKeyDown() then
			-- unused
		elseif IsAltKeyDown() then
			-- toggle monitoring
			BrokerRaidFinder:TriggerAction("toggle")
		else
			-- show finder log
			BrokerRaidFinder:TriggerAction("show")
		end
	end
end

function BrokerRaidFinder:TriggerAction(action, args)
	if action == "menu" then
		-- open options menu
		InterfaceOptionsFrame_OpenToCategory(FULLNAME)
	elseif action == "version" then
		-- print version information
		BrokerRaidFinder:PrintVersionInfo()
	elseif action == "debug" then
		if args[2] == "on" then
			self:Output("debug mode turned on")
			self.debug = true
		end
		if args[2] == "off" then
			self:Output("debug mode turned off")
			self.debug = false
		end
	elseif action == "toggle" then
		self:ToggleMonitoring()
	elseif action == "on" then
		self:ActivateMonitoring(true)
	elseif action == "off" then
		self:ActivateMonitoring(false)
	elseif action == "show" then
		self:ShowLogWindow()
	elseif action == "friends" then
		if next(self.comm_friends) then
			for player, _ in pairs(self.comm_friends) do
				self:Output("Friend: " .. player)
			end
		else
			self:Output("No friends.")
		end
		if next(self.comm_bnetfriends) then
			for player, _ in pairs(self.comm_bnetfriends) do
				self:Output("BNetFriend: " .. player)
			end
		else
			self:Output("No bnetfriends.")
		end
	elseif action == "remote" then
		self:Output("Partners:")
		for partner, _ in pairs(self.partners) do
			self:Output(tostring(partner))
		end
		self:Output("Clients:")
		for client, _ in pairs(self.clients) do
			self:Output(tostring(client))
			self:Output("Instances:")
			for instance, clients in pairs(self.comm_clientconfig) do
				for c, _ in pairs(clients) do
					if c == client then
						self:Output(instance)
					end
				end
			end
		end
		if self.server then
			self:Output("Server: " .. tostring(self.server))
		else
			self:Output("Server: none")
		end
		self:Output("Blacklisted:")
		for blacklisted, _ in pairs(self.comm_blacklist) do
			self:Output(tostring(blacklisted))
		end
		self:Output("Comm quality: " .. tostring(self.comm_quality))
		self:Output("Comm numclients: " .. tostring(self.comm_numclients))
		self:Output("Numclients: " .. tostring(self.numclients))
	elseif action == "saved" then
		if next(self.player_saved_raids) then
			for instance, _ in pairs(self.player_saved_raids) do
				self:Output(instance)
			end
		else
			self:Output("No saved raids.")
		end
	else -- if action == "help" then
		-- display help
		self:Output(L["Usage:"])
		self:Output(L["/braidfinder arg"])
		self:Output(L["/brfind arg"])
		self:Output(L["Args:"])
		self:Output(L["version - display version information"])
		self:Output(L["menu - display options menu"])
		self:Output(L["on - activate monitoring"])
		self:Output(L["off - deactivate monitoring"])
		self:Output(L["show - show log window"])
		self:Output(L["help - display this help"])
	end
end

-- update functions
function BrokerRaidFinder:Update()
	self:UpdateData()
	self:UpdateLabel()
end

function BrokerRaidFinder:UpdateData()
	self:UpdateLogList()

	if self.tainted_data == false then
		return
	end

	local players = {}
	local instances = {}
	
	local now = time()
		
	self.matches_latest = {}
	self.match_oldest   = nil
	
	for _, match in pairs(self.db.faction.matches) do
		if match.timestamp + self.db.profile.timeFrame*60 < now then
			break
		end

		if self:IsMonitored(match.instance) and not self:ExcludeAsSaved(match.instance) then
			players[match.player] = true
			instances[match.instance] = true

			if not self.matches_latest[match.instance] then
				local last = {
					player    = match.player,
					timestamp = match.timestamp,
					players   = { }
				}
				self.matches_latest[match.instance] = last
			end

			self.matches_latest[match.instance].players[match.player] = true
			
			self.match_oldest = match
		end
	end
	
	self.matched_players   = 0
	
	for _ in pairs(players) do
		self.matched_players = self.matched_players + 1
	end

	self.matched_instances = 0
	
	for _ in pairs(instances) do
		self.matched_instances = self.matched_instances + 1
	end	

	for _, match in pairs(self.matches_latest) do
		local count = 0
		
		for _ in pairs(match.players) do
			count = count + 1
		end
		
		match.players = count
	end
		
	self.tainted_data = false
end

function BrokerRaidFinder:UpdateLabel()
	local text = L["I: "] .. tostring(self.matched_instances) .. " / " .. L["P: "] .. tostring(self.matched_players)
	
	if not self.db.profile.monitoringActive or not next(self.db.profile.monitored) then
		text = GrayOut(text)
	elseif self.matched_instances == 0 then
		text = Red(text)
	else
		text = Green(text)
	end
	
	if self.server then
		text = text .. " " .. Green("!")
	end

	dataobj.text = text
end

function BrokerRaidFinder:UpdateIcon()
    if self.db.profile.monitoringActive then
		dataobj.icon = "Interface\\Addons\\"..DIRNAME.."\\icon"
    else
		dataobj.icon = "Interface\\Addons\\"..DIRNAME.."\\icondisabled"
    end
	
	self:SetMinimapIcon(self.db.profile.monitoringActive)
end

function BrokerRaidFinder:UpdateLogList()
	self.filtered_matches = {}

	local now   = time()
	local found = false 
	
	for _, match in pairs(self.db.faction.matches) do
		-- only show if within specified timeframe
		if match.timestamp + self.db.profile.timeFrame*60 < now then
			break
		end
				
		-- check source filter
		local validSource = false
		
		if self.sourcefilter == FILTERSRCOPT_ALL then
			validSource = true
		elseif self.sourcefilter == FILTERSRCOPT_SELF then
			if not match.source and match.char == self.PlayerName then
				validSource = true
			end
		elseif self.sourcefilter == FILTERSRCOPT_ALTS then
			if not match.source and match.char ~= self.PlayerName then
				validSource = true
			end
		elseif self.sourcefilter == FILTERSRCOPT_REMOTE then
			if match.source then
				validSource = true
			end
		end
		
		-- check instance filter
		local validInstance = false

		if self.instancefilter == FILTERINSTOPT_ALL then
			validInstance = true
		else
			if match.instance == self.instancefilter then
				validInstance = true
			end
		end
		
		-- insert into filtered matches
		if validSource and validInstance then
			tinsert(self.filtered_matches, match)
			
			if match == self.selected_match then
				found = true
			end
		end
	end
	
	if not found then
		-- reset selection
		self.selected_match = nil

		self:DisplaySelectedMessage()
	end
	
	MatchListUpdate()
end

-- update cached data function
function BrokerRaidFinder:UpdateRaidInfo()
	self.player_saved_raids = {}

	local numSaved = GetNumSavedInstances()
	
	-- we do not distinguish between different difficulties
	for i = 1, numSaved do
		local instanceName, instanceID, instanceReset, instanceDifficulty, locked, extended, instanceIDMostSig, isRaid, maxPlayers, difficultyName, maxBosses, defeatedBosses = GetSavedInstanceInfo(i)
		
		if isRaid and locked then
			instance = self:FuzzySearchInstance(instanceName)

			if instance then
				self.player_saved_raids[instance] = true
			end
		end
	end
end

-- since we need to map the loclized name GetSavedInstanceInfo returns to our internal non-localized instance name
-- we perform a very simple matching algorithm to do the task
-- if pattern contains all of the significant words (longer than 3 chars) we assume we found the match
function BrokerRaidFinder:FuzzySearchInstance(pattern)
	if not pattern then
		return
	end
	
	pattern = strlower(pattern)

	for name, _ in pairs(brf_DefaultInstanceKeywords) do
		local localized = LibBabbleZone[name]
		localized = strlower(localized)
		
		local matched = true
		
		for word in string.gmatch(localized, "[^%s%p%c]+") do
		    -- significant length of a word is 4 characters
			-- not sure if that works well with all possible localizations
			if string.len(word) > 3 then
				matched = matched and strfind(pattern, word)
			end
		end
		
		if matched then
			return name
		end
	end
	
	return 
end

function BrokerRaidFinder:UpdateRaidUnits()
	self.raidNames = {}
	
	if GetNumRaidMembers() > 0 then
		for i = 1, 40, 1 do
			local unit = "raid"..i
			local name = UnitName(unit)
			
			if name then
				self.raidNames[name] = true
			end
		end
	elseif GetNumPartyMembers() > 0 then
		for i = 1, 4, 1 do
			local unit = "party"..i
			local name = UnitName(unit)
			
			if name then
				self.raidNames[name] = true
			end
		end
	end
end

-- update keyword lists
function BrokerRaidFinder:SetDefaultInstanceKeywords(init)
	for instance, _ in pairs(brf_DefaultInstanceKeywords) do
		self:SetDefaultKeywords(instance, init)
	end
end

function BrokerRaidFinder:SetDefaultKeywords(instance, init)
	if init then
		defaults.global.instanceKeywords[instance] = brf_DefaultInstanceKeywords[instance]
	else
		self.db.global.instanceKeywords[instance] = brf_DefaultInstanceKeywords[instance]
	end
end

function BrokerRaidFinder:UpdateAllKeywords()
	self:UpdateLFGKeywords()

	for instance, _ in pairs(self.db.global.instanceKeywords) do
		self:UpdateInstanceKeywords(instance)
	end
end

function BrokerRaidFinder:UpdateLFGKeywords()
	self.lfgKeywordList = self:GetKeywordList(self.db.global.lfgKeywords)
end

function BrokerRaidFinder:UpdateInstanceKeywords(instance)
	if not instance then
		return
	end
	
	self.instanceKeywordLists[instance] = self:GetKeywordList(self.db.global.instanceKeywords[instance])	
end

function BrokerRaidFinder:GetKeywordList(keywords)
	if not keywords then
		return {}
	end

	local list =  {}
	
	-- get all words separated by comma and trim left and right
	for word in string.gmatch(keywords, "[^,]+") do 
		word = word:gsub("^%s*(.-)%s*$", "%1")
		if word ~= "" then
			list[#list+1] = strlower(word)
		end
	end
	
	-- sort list by keyword length, shortest first
	table_sort(list, keyword_lt)
	
	return list
end

-- tooltip
function BrokerRaidFinder:DrawTooltip()
	tooltip:Hide()
	tooltip:Clear()

	local now = time()
	
	-- module header
	local lineNum = tooltip:AddHeader( " " )
	tooltip:SetCell( lineNum, 1, White( FULLNAME ), "CENTER", tooltip:GetColumnCount() )
	
	local ctrlKeyPressed = IsControlKeyDown()
	
	-- events header
	if ctrlKeyPressed or self.db.profile.monitoringActive then
		-- for all extensions
		for extension, instances in pairs(brf_Instances) do
			local header_present = false
			
			-- for all instances
			for _, instance in pairs(instances) do
				local monitored = self:IsMonitored(instance)
				
				if ctrlKeyPressed or (self:IsMonitored(instance) and not self:ExcludeAsSaved(instance)) then 
					local saved = self:IsSaved(instance)
				
					if not header_present then
						-- set extension name as header
						tooltip:AddLine( " " )
						lineNum = tooltip:AddLine( " " )
						tooltip:SetCell( lineNum, 1, Orange( L[brf_Extensions[extension]] ), "LEFT", tooltip:GetColumnCount() )
						
						header_present = true
					end

					lineNum = tooltip:AddLine( " " )
					
					if self.matches_latest[instance] then
						-- instance
						if saved then
							tooltip:SetCell( lineNum, 1, Red( LibBabbleZone[instance] ) .. " ", "LEFT")
						else
							tooltip:SetCell( lineNum, 1, Green( LibBabbleZone[instance] ) .. " ", "LEFT")
						end

						local match = self.matches_latest[instance]
						
						-- last player posting
						tooltip:SetCell( lineNum, 2, self:ColorizeChar( match.player ) .. White(" (of " .. tostring(match.players) .. ") ") , "LEFT")
						
						-- time since posting
						local timestamp = now - match.timestamp
						local timestring = ""
						if Crayon then
							local timeframe = self.db.profile.timeFrame * 60
							timestring = "|cff"..Crayon:GetThresholdHexColor(timestamp, timeframe, timeframe * 0.75, timeframe * 0.5, timeframe * 0.25, 1) .. format_time(timestamp) .. "|r"
						else
							timestring = format_time(timestamp)
						end				
						
						tooltip:SetCell( lineNum, 3, timestring, "RIGHT")
					else
						-- instance
						if saved then
							tooltip:SetCell( lineNum, 1, Red( LibBabbleZone[instance] ) .. " ", "LEFT")
						elseif monitored then
							tooltip:SetCell( lineNum, 1, Blueish( LibBabbleZone[instance] ) .. " ", "LEFT")
						else
							tooltip:SetCell( lineNum, 1, GrayOut( LibBabbleZone[instance] ) .. " ", "LEFT")
						end
						
						-- no player
						tooltip:SetCell( lineNum, 2, GrayOut( "--" ), "LEFT")
						
						-- no time
						tooltip:SetCell( lineNum, 3, GrayOut( "--" ), "RIGHT")
					end
					
					tooltip:SetCellScript( lineNum, 1, "OnMouseDown", BRF_ToggleInstance, instance)					
				end			
			end	
		end
		
		if self.server then
			tooltip:AddLine( " " )
			lineNum = tooltip:AddLine( " " )
			tooltip:SetCell( lineNum, 1, Green( L["Remote client is monitoring."] ), "LEFT", tooltip:GetColumnCount() )
		end
	else
		lineNum = tooltip:AddLine( " " )
		tooltip:SetCell( lineNum, 1, Red( L["Monitoring is switched off."] ), "LEFT", tooltip:GetColumnCount() )
	end
	
	-- hint to show options
	if not self.db.profile.hideHint then
		tooltip:AddLine( " " )
		lineNum = tooltip:AddLine( " " )
		tooltip:SetCell( lineNum, 1, Brownish( L["Left-Click"] .. ":") .. Yellow( " " .. L["Show logging window."] ), "LEFT", tooltip:GetColumnCount() )
		lineNum = tooltip:AddLine( " " )
		tooltip:SetCell( lineNum, 1, Brownish( L["Alt-Click"] .. ":") .. Yellow( " " .. L["Enable/disable monitoring."] ), "LEFT", tooltip:GetColumnCount() )
		lineNum = tooltip:AddLine( " " )
		tooltip:SetCell( lineNum, 1, Brownish( L["Right-Click"] .. ":") .. Yellow( " " .. L["Open option menu."] ), "LEFT", tooltip:GetColumnCount() )
		lineNum = tooltip:AddLine( " " )
		tooltip:SetCell( lineNum, 1, Brownish( L["Left-Click Tip"] .. ":") .. Yellow( " " .. L["Click on instance name to toggle monitoring."] ), "LEFT", tooltip:GetColumnCount() )
		lineNum = tooltip:AddLine( " " )
		tooltip:SetCell( lineNum, 1, Brownish( L["Ctrl-Hover Tip"] .. ":") .. Yellow( " " .. L["Press Ctrl while opening tooltip shows all instances."] ), "LEFT", tooltip:GetColumnCount() )
	end
	
end

-- event handlers
function BrokerRaidFinder:SetupEventHandlers()
	if self.db.profile.monitoringActive == true then
		self:RegisterEvent("CHAT_MSG_CHANNEL")
	else
		self:UnregisterEvent("CHAT_MSG_CHANNEL")	
	end

	self:SetupGuildChatEventHandlers()
end

function BrokerRaidFinder:SetupGuildChatEventHandlers()
	if self.db.profile.monitoringActive == true and self.db.profile.monitorGuild == true then
		self:RegisterEvent("CHAT_MSG_GUILD")
		self:RegisterEvent("CHAT_MSG_OFFICER")	
	else
		self:UnregisterEvent("CHAT_MSG_GUILD")
		self:UnregisterEvent("CHAT_MSG_OFFICER")	
	end
end

function BrokerRaidFinder:ChatEventHandler(message, author, event, localOnly)
	self:HandleMessage(message, author, localOnly)
end

function BrokerRaidFinder:CHAT_MSG_CHANNEL(event, message, author, arg3, arg4, arg5, arg6, arg7, id)
	if self:IsMonitoredChannel(id) then
		self:HandleMessage(message, author)
	end
end

function BrokerRaidFinder:CHAT_MSG_GUILD(event, message, author)
	self:ChatEventHandler(message, author, event, true)
end

function BrokerRaidFinder:CHAT_MSG_OFFICER(event, message, author)
	self:ChatEventHandler(message, author, event, true)
end

-- saved raid info events
function BrokerRaidFinder:UPDATE_INSTANCE_INFO()
  self:UpdateRaidInfo()
end

function BrokerRaidFinder:RAID_INSTANCE_WELCOME()
	RequestRaidInfo()
end

function BrokerRaidFinder:CHAT_MSG_SYSTEM(msg)
	if tostring(msg) == _G["INSTANCE_SAVED"] then
		RequestRaidInfo()
	end
end

-- log window
function BrokerRaidFinder:ShowLogWindow()
	self.logwindow:Show()
--	if (not self.logwindow:IsShown()) then
--		self.logwindowrequested = true
--		self:UpdateLogWindow()
--		self.logwindow:Show()
--		self.logwindowrequested = nil
--	end
end

function BrokerRaidFinder:HandleInstanceFilter(value)
	if not value or value == self.instancefilter then
		return
	end
	
	self.instancefilter = value
		
	self:UpdateLogList()
end

function BrokerRaidFinder:HandleSourceFilter(value)
	if not value or value == self.sourcefilter then
		return
	end
	
	self.sourcefilter = value
	
	self:UpdateLogList()
end

function BrokerRaidFinder:SetSelectedEntry(selected)
	if not selected then
		return
	end
	
	-- deselect old match
	for i=1, MAX_LIST_SIZE do
		local entry = self.matchlistframe.entries[i]

		if entry.data == self.selected_match then
			entry:SetBackdropColor(0, 0, 0, .33)
		end
	end
	
	-- really needed? if we click the frame OnEnter should have colored it properly before anyway
	selected:SetBackdropColor(1, 1, 1, 1)
	
	self.selected_match = selected.data
	
	self:DisplaySelectedMessage()
end

function BrokerRaidFinder:DisplaySelectedMessage()
	if self.selected_match then
		local match = self.selected_match
		local text = LibBabbleZone[match.instance] .. " - [" .. format_time(match.timestamp + self.time_offset) .. "] " .. self:ColorizeChar(match.player) .. ": " .. match.message
		
		self.messageframe.text:SetText(text)
		
		-- enable whisper button
		self.whisperbutton:Enable()		
	else
		self.messageframe.text:SetText("")
		
		-- disable whisper button
		self.whisperbutton:Disable()		
	end
end

-- user functions
function BrokerRaidFinder:PrintVersionInfo()
    self:Output(L["Version"] .. " " .. White(GetAddOnMetadata(DIRNAME, "Version")))
end

-- utilities
function BrokerRaidFinder:Output(msg)
	if ( msg ~= nil and DEFAULT_CHAT_FRAME ) then
		DEFAULT_CHAT_FRAME:AddMessage( MODNAME..": "..msg, 0.6, 1.0, 1.0 )
	end
end

function BrokerRaidFinder:IsMonitoredChannel(id)
	return monitored_channels[id] == true
end

function BrokerRaidFinder:ColorizeChar(name)
	if not name then
		return
	end
	
	local class = self.classCache[name]
	
	if not class then
		local info = self:UserInfo(name, { callback = 'UnitInfoCallback' })
		
		if info and info.Class then
			class = gsub(strupper(info.Class), " ", "")
			self.classCache[name] = class
		end
	end
	
	if class then
		local colortable = _G["CUSTOM_CLASS_COLORS"] or _G["RAID_CLASS_COLORS"]
		local color = colortable[class]
		return "|cff" .. string.format("%02x%02x%02x", color.r*255, color.g*255, color.b*255) .. name .. "|r"
	else
		return GrayOut( name )
	end
end

function BrokerRaidFinder:UnitInfoCallback(user, time)
	if user and user.Class then
		self.classCache[user.Name] = gsub(strupper(user.Class), " ", "")
	end		
end

function BrokerRaidFinder:IsMonitored(instance)
	return self.db.profile.monitored[instance] == true
end

function BrokerRaidFinder:IsSaved(instance)
	return self.player_saved_raids[instance] == true
end

function BrokerRaidFinder:ExcludeAsSaved(instance)
	if self.db.profile.excludeSavedRaids then
		return self.player_saved_raids[instance] == true
	else
		return false
	end
end

function BrokerRaidFinder:PlaySoundFile()
	local sound = soundFiles[self.db.profile.notificationSound]

	if ((string.sub(sound, -4) == ".wav") or (string.sub(sound, -4) == ".mp3")) then
		PlaySoundFile(sound);
	else
		PlaySound(sound);
	end
end

function BrokerRaidFinder:IsInRaid(name)
	return self.raidNames[name] ~= nil
end

function BrokerRaidFinder:WhisperAuthor()
	if not self.selected_match then
		return
	end

	DEFAULT_CHAT_FRAME.editBox:Show()
	DEFAULT_CHAT_FRAME.editBox:SetText("/whisper " .. self.selected_match.player .. " ")
end

-- testing
function BrokerRaidFinder:Debug(msg)
	if self.debug then
		if ( msg ~= nil and DEFAULT_CHAT_FRAME ) then
			DEFAULT_CHAT_FRAME:AddMessage( MODNAME .. " (dbg): " .. msg, 1.0, 0.37, 0.37 )
		end
	end
end

-- data handling
function BrokerRaidFinder:HandleMessage(msg, author, localOnly)
	local match, instance = self:ProcessMessage(msg, author)
	
	if not localOnly then
		self:ProcessMessageForClients(msg, author, match, instance)
	end
end

function BrokerRaidFinder:ProcessMessage(msg, author, source)
	if self.db.profile.monitoringActive == false or
	   author == self.PlayerName or  
	   self:IsInRaid(author) then
		return
	end

	local lmsg = strlower(msg)
	local hasMatch = false
	
	-- check for 'looking for' keywords
	for _, keyword in pairs(self.lfgKeywordList) do
		if strfind(lmsg, keyword) then
			hasMatch = true
			break
		end
	end
	
	if hasMatch == false then
		return
	end
	
	-- after 'looking for' match we need to find an instance match
	local matchedkeyword  = nil
	local matchedinstance = nil
	
	-- for all extensions
	for extension, instances in pairs(brf_Instances) do
		-- for all instances
		for _, instance in pairs(instances) do
		
			if self:IsMonitored(instance) then 
				-- for all keywords
				local keywords = self.instanceKeywordLists[instance] or {}
				
				for _, keyword in pairs(keywords) do
					if strfind(lmsg, keyword) then
						matchedkeyword  = keyword
						matchedinstance = instance
						
						local match = {
							timestamp = time(),
							char      = self.PlayerName,
							instance  = instance,
							player    = author,
							message   = msg,
							source    = source
						}
						
						self:InsertMatch(match)
						
						-- request user info, so we got it when we need it
						self:UserInfo(author, { callback = 'UnitInfoCallback' })
						
						-- check notification timeout
						local donotify = true
						
						if not self.notification_timeouts[author] then
							self.notification_timeouts[author] = {}
						end
						
						if self.notification_timeouts[author][instance] and self.notification_timeouts[author][instance] + self.db.profile.notificationTimeout*60 > match.timestamp then
							donotify = false
						end						

						-- notification only if last match wasn't by same player to prevent spamming					
						if donotify then
							if self.db.profile.notifyText then
								self:Output("Found new match for " .. instance .." by player " .. author .. ".")
							end

							if self.db.profile.notifySound then
								self:PlaySoundFile()
							end
							
							-- update notification timeout timestamp
							self.notification_timeouts[author][instance] = match.timestamp
						end
						
						self:Update()						
						break
					end
				end
			end
		end
	end
	
	return matchedkeyword, matchedinstance
end

function BrokerRaidFinder:InsertMatch(match)
	if match == nil then
		return
	end
	
	-- remove last entry by author to the very same instance	
	for index, entry in pairs(self.db.faction.matches) do
		if entry.instance == match.instance and
		   entry.player == match.player then
			tremove(self.db.faction.matches, index)
			
			break
		end
	end
	
	-- insert at the beginning
	tinsert(self.db.faction.matches, 1, match)

	self.tainted_data = true	
end					

function BrokerRaidFinder:ManageHistory()
	local now           = time()
	local oldestvalid   = now - GTIMERANGE*60
	local oldestframe   = now - self.db.profile.timeFrame*60
	
	-- remove all matches outside global maximum time range
	while #self.db.faction.matches ~= 0 and self.db.faction.matches[#self.db.faction.matches].timestamp < oldestvalid do
		tremove(self.db.faction.matches)
		self.tainted_data = true
	end
	
	if self.match_oldest and self.match_oldest.timestamp < oldestframe then
		self.tainted_data = true
	end

	if self.tainted_data then
		self:Update()
	end
end

-- settings
function BrokerRaidFinder:ActivateMonitoring(activate)
	if not activate then
		activate = false
	end

	if self.db.profile.monitoringActive ~= activate then
		self:ToggleMonitoring()
		self:UpdateLabel()
	end
end

function BrokerRaidFinder:ToggleMonitoring()
	self.db.profile.monitoringActive = not self.db.profile.monitoringActive
	
	-- setup event handlers
	self:SetupEventHandlers()
	
	self:UpdateIcon()
	self:Update()	
end

function BrokerRaidFinder:ToggleMonitorGuild()
	self.db.profile.monitorGuild = not self.db.profile.monitorGuild
	
	-- setup event handlers
	self:SetupGuildChatEventHandlers()
end

function BrokerRaidFinder:ToggleSavedRaids()
	self.db.profile.excludeSavedRaids = not self.db.profile.excludeSavedRaids
	
	self.tainted_data = true
	
	self:Update()
end

function BrokerRaidFinder:ToggleAddonCommunication()
	self.db.profile.addonCommunication = not self.db.profile.addonCommunication
	
	-- handle comm
	self:PrepareCommunication()
end

function BrokerRaidFinder:ToggleInstance(instance)
	if not brf_DefaultInstanceKeywords[instance] then
		return
	end

	if self.db.profile.monitored[instance] == true then
		self.db.profile.monitored[instance] = nil
	else
		self.db.profile.monitored[instance] = true
	end
	
	self.tainted_data = true
	
	self:UpdateConfigData()

	self:Update()
end

